from spydrnet.ir.outerpin import OuterPin


class Comparer:

    def __init__(self, ir_orig, ir_composer):
        assert ir_orig and ir_composer, "Both orig_netlist and composer netlist must be specified"
        self.ir_orig = ir_orig
        self.ir_composer = ir_composer

    def run(self):
        self.compare()

    def compare(self):
        assert self.get_identifier(self.ir_orig) == self.get_identifier(self.ir_composer), \
            "Environments do not have the same identifier"
        assert self.get_original_identifier(self.ir_orig) == self.get_original_identifier(self.ir_composer), \
            "Environments do not have the same original identifier"
        self.compare_instances(self.ir_orig.top_instance, self.ir_composer.top_instance)
        assert len(self.ir_orig.libraries) == len(self.ir_composer.libraries), \
            "Environments do not have the same number of libraries"
        for orig_library, composer_library in zip(self.ir_orig.libraries, self.ir_composer.libraries):
            self.compare_libraries(orig_library, composer_library)

    def compare_libraries(self, library_orig, library_composer):
        assert self.get_identifier(library_orig) == self.get_identifier(library_composer), \
            "Libraries do not have the same identifier"
        assert self.get_original_identifier(library_orig) == self.get_original_identifier(library_composer), \
            "Libraries do not have the same original identifier"
        assert len(library_orig.definitions) == len(library_composer.definitions), \
            "Libraries do not have the same amount of definitions"
        for orig_definition, composer_definition in zip(library_orig.definitions, library_composer.definitions):
            self.compare_definition(orig_definition, composer_definition)

    def compare_definition(self, definition_orig, definition_composer, check_identifier=True):
        if check_identifier:
            assert self.get_identifier(definition_orig) == self.get_identifier(definition_composer), \
                "Definitions do not have the same identifier"
            assert self.get_original_identifier(definition_orig) == self.get_original_identifier(definition_composer), \
                "Definitions do not have the same original identifier"

        assert len(definition_orig.ports) == len(definition_composer.ports), \
            "Definitions do not have the same number of ports"
        for orig_port, composer_port in zip(definition_orig.ports, definition_composer.ports):
            self.compare_ports(orig_port, composer_port)

        assert len(definition_orig.cables) == len(definition_composer.cables), \
            "Definitions do not have the same number of cables"
        for orig_cable, composer_cable in zip(definition_orig.cables, definition_composer.cables):
            self.compare_cables(orig_cable, orig_cable)

        assert len(definition_orig.children) == len(definition_composer.children), \
            "Definitions do not have the same number of instances"
        for orig_instance, composer_cable in zip(definition_orig.children, definition_composer.children):
            self.compare_instances(orig_instance, composer_cable)

    def compare_cables(self, cable_orig, cable_composer):
        assert self.get_identifier(cable_orig) == self.get_identifier(cable_composer), \
            "Cables do not have the same identifier"
        assert self.get_original_identifier(cable_orig) == self.get_original_identifier(cable_composer), \
            "Cables do not have the same original identifier"
        assert len(cable_orig.wires) == len(cable_composer.wires), \
            "Cables do not have the same number of wires"
        for orig_wire, composer_wire in zip(cable_orig.wires, cable_composer.wires):
            for orig_pin, composer_pin in zip(orig_wire.pins, composer_wire.pins):
                assert type(orig_pin) == type(composer_pin), \
                    "Environments do not have the same number of libraries"
                if isinstance(orig_pin, OuterPin):
                    self.compare_outer_pins(orig_pin, composer_pin)
                else:
                    self.compare_inner_pins(orig_pin, composer_pin)

    def compare_outer_pins(self, pin_orig, pin_composer):
        assert pin_orig.instance.reference == pin_orig.inner_pin.port.definition and \
                pin_composer.instance.reference == pin_composer.inner_pin.port.definition, \
                "DRC failure, outer pin instance reference on associated pin definition not the same"
        assert self.are_instances_equivalent(pin_orig.instance, pin_composer.instance), \
            "Net does not connect to a pin on the same instance"
        assert self.are_inner_pins_equivalent(pin_orig.inner_pin, pin_orig.inner_pin), \
            "Net does not connect the same pin"

    def compare_inner_pins(self, pin_orig, pin_composer):
        assert self.are_inner_pins_equivalent(pin_orig, pin_composer), \
            "Net does not connect the same pin"

    def are_instances_equivalent(self, orig_instance, composer_instance):
        assert self.get_identifier(orig_instance) == self.get_identifier(composer_instance) and \
            self.get_identifier(orig_instance.reference) == self.get_identifier(composer_instance.reference) and \
            self.get_identifier(orig_instance.reference.library) == \
               self.get_identifier(composer_instance.reference.library) and \
            self.get_identifier(orig_instance.parent) == self.get_identifier(composer_instance.parent) and \
            self.get_identifier(orig_instance.parent.library) == \
                self.get_identifier(composer_instance.parent.library), \
            'Instances are not equivalent'
        return True

    def are_inner_pins_equivalent(self, orig_pin, composer_pin):
        assert orig_pin.port.pins.index(orig_pin) == composer_pin.port.pins.index(composer_pin), \
            "Pin indices do not match"
        assert self.get_identifier(orig_pin.port) == self.get_identifier(composer_pin.port) and \
            self.get_identifier(orig_pin.port.definition) == self.get_identifier(composer_pin.port.definition) and \
            self.get_identifier(orig_pin.port.definition.library) == \
                self.get_identifier(composer_pin.port.definition.library), \
            "Pins are not from equivalent ports"
        return True

    def compare_ports(self, port_orig, port_composer):
        assert self.get_identifier(port_orig) == self.get_identifier(port_composer), \
            "Ports do not have the same identifier"
        assert self.get_original_identifier(port_orig) == self.get_original_identifier(port_composer), \
            "Ports do not have the same original identifier"

        assert port_orig.direction == port_composer.direction, \
            "Ports are not facing the same direction"

        assert port_orig.is_array == port_composer.is_array, \
            "Ports Array mismatch"

        assert len(port_orig.pins) > 0 and len(port_composer.pins) > 0, \
            "DRC failure, ports should have at least one pin"

        assert len(port_orig.pins) == len(port_composer.pins), \
            "Ports do not have the same number of pins"

        assert self.get_identifier(port_orig.definition) == self.get_identifier(port_composer.definition) and \
            self.get_identifier(port_orig.definition.library) == \
               self.get_identifier(port_composer.definition.library), \
            "Ports do not belong to an equivalent definition"

    def compare_instances(self, instances_orig, instances_composer):
        instances_orig_identifier = self.get_identifier(instances_orig)
        instances_composer_identifier = self.get_identifier(instances_composer)
        assert instances_orig_identifier == instances_composer_identifier, \
            "Instances do not have the same identifier, orig \"{}\" " \
            "composer \"{}\"".format(instances_orig_identifier, instances_composer_identifier)

        instances_orig_original_identifier = self.get_original_identifier(instances_orig)
        instances_composer_original_identifier = self.get_original_identifier(instances_composer)
        assert instances_orig_original_identifier == instances_composer_original_identifier, \
            "Instances do not have the same original identifier, orig " \
            "\"{}\" composer \"{}\"".format(instances_orig_original_identifier, instances_composer_original_identifier)

        assert self.get_identifier(instances_orig.reference) == self.get_identifier(instances_composer.reference) and \
            self.get_identifier(instances_orig.reference.library) == \
               self.get_identifier(instances_composer.reference.library), \
            "Instances do not have the same reference definition."

        if "EDIF.properties" in instances_orig:
            assert "EDIF.properties" in instances_composer, "Composer is missing properties"
            properties_orig = instances_orig["EDIF.properties"]
            properties_composer = instances_composer["EDIF.properties"]
            for x in range(len(properties_orig)):
                for key, value in properties_orig[x].items():
                    assert properties_orig[x][key] == properties_composer[x][key], \
                        "Instances do not have the same properties"

    @staticmethod
    def get_identifier(obj):
        if "EDIF.identifier" in obj:
            return obj["EDIF.identifier"]

    @staticmethod
    def get_original_identifier(obj):
        if "EDIF.original_identifier" in obj:
            return obj["EDIF.original_identifier"]



import json
from spydrnet.ir import *
import inspect  # used for debug.
from datetime import datetime


class ComposeEdif:
    """class that creates an edif file from our IR"""

    def __init__(self):
        # self.filename = "../../_data_/json_edif/out.json"
        self._output_ = None
        self._data_ = None
        self._lisp_depth_ = 0
        self.test = 0

    def run(self, ir=None, file_out="out.edf"):
        """
        compose an edif file from the IR in either file or object form and _output_ it to a file
        **currently only object form will work read the json into an object first**

        Keyword Arguments:
        ir -- the object(environment) or file(json) to be composed to edif (default None)
                        **default should be changed to in.ir**
        file_out -- the path and name of the file to which the edif will be written (default "out.edf")
        """
        self.output_filename = file_out
        self._data_ = ir
        if (isinstance(ir, str)):
            self.filename = ir
            self._read_data_()  # only needed if we start to accept the json format files
            print("currently json files are unsupported! read into an object first.")
        else:
            self._data_ = ir
        self._open_output_()
        self._output_environment_()
        self._close_output_()

    def _read_data_(self):
        """read _data_ in from a json ir file and store it in _data_"""
        with open(self.filename) as fi:
            self._data_ = json.load(fi)

    def _open_output_(self):
        self._output_ = open(self.output_filename, 'w', buffering=1)

    def _output_environment_(self):
        self._lisp_increment_()
        self._output_.write("edif ")
        self._output_name_of_object_(self._data_)
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("edifversion 2 0 0")
        self._lisp_decrement_()
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("edifLevel 0")
        self._lisp_decrement_()
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("keywordmap ")
        self._lisp_increment_()
        self._output_.write("keywordlevel 0")
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._output_status_()
        self._new_line_()
        for library in self._data_.libraries:
            self._output_library_(library)

        self._lisp_increment_()
        self._output_.write("design ")
        self._output_name_of_object_(self._data_.top_instance)
        self._new_line_()
        self._lisp_increment_()
        self._output_.write("cellref ")
        test = self._data_.top_instance
        self._output_.write(self._data_.top_instance.reference['EDIF.identifier'])
        self._lisp_increment_()
        self._output_.write("libraryref ")
        self._output_.write(self._data_.top_instance.reference.library['EDIF.identifier'])
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()
        self._lisp_decrement_()
        self._new_line_()

        self._lisp_decrement_()
        
        assert self._lisp_depth_ == 0, "There was an error with parenthesis matching off by " + str(self._lisp_depth_)

    def _output_status_(self):
        self._new_line_()
        self._lisp_increment_()
        self._output_.write("status")
        self._new_line_()
        self._lisp_increment_()
        self._output_.write("written")
        self._new_line_()
        self._lisp_increment_()
        self._output_.write("timeStamp ")
        now = datetime.now()
        self._output_.write("{}".format(now.strftime("%Y %m %d %H %M %S")))
        self._lisp_decrement_()
        self._new_line_()
        if 'EDIF.status.written.program' in self._data_:
            self._lisp_increment_()
            self._output_.write("program ")
            self._output_.write('"{}" '.format(self._data_['EDIF.status.written.program']))
            if 'EDIF.status.written.program.version' in self._data_:
                self._lisp_increment_()
                self._output_.write("version ")
                self._output_.write('"{}"'.format(self._data_['EDIF.status.written.program.version']))
                self._lisp_decrement_()
            self._lisp_decrement_()
        self._new_line_()
        self._lisp_increment_()
        self._output_.write('comment "Built by \'BYU spydrnet tool\'"')
        self._lisp_decrement_()
        self._new_line_()
        self._lisp_decrement_()
        self._new_line_()
        self._lisp_decrement_()

    def _output_library_(self, library):
        self._lisp_increment_()
        self._output_.write("Library ")
        self._output_name_of_object_(library)
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("edifLevel 0")
        self._lisp_decrement_()
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("technology ")
        self._lisp_increment_()
        self._output_.write("numberDefinition ")
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()

        for definition in library.definitions:
            self._output_definition_(definition)

        self._lisp_decrement_()
        self._new_line_()

    def _output_name_of_object_(self, obj):
        if 'EDIF.original_identifier' in obj:
            self._lisp_increment_()
            self._output_.write("rename ")
            self._output_.write(obj["EDIF.identifier"])
            self._output_.write(" \"" + obj['EDIF.original_identifier'] + "\"")
            self._lisp_decrement_()
        else:
            # print(obj)
            # print(vars(obj))
            # self._output_.write(obj.name)
            self._output_.write(obj["EDIF.identifier"])

    def _output_definition_(self, definition):
        self._lisp_increment_()
        self._output_.write("Cell ")
        self._output_name_of_object_(definition)
        self._output_.write(" ")
        self._lisp_increment_()
        self._output_.write("celltype GENERIC")
        self._lisp_decrement_()
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("view netlist ")
        self._lisp_increment_()
        self._output_.write("viewtype NETLIST")
        self._lisp_decrement_()
        self._new_line_()

        self._lisp_increment_()
        self._output_.write("interface")
        self._new_line_()
        for port in definition.ports:
            self._output_port_(port)
        self._lisp_decrement_()
        self._new_line_()

        if len(definition.children) + len(definition.cables) > 0:
            self._lisp_increment_()
            self._output_.write("contents")
            self._new_line_()

            for instance in definition.children:
                self._output_instance_(instance)
            for cable in definition.cables:
                self._output_cable_(cable)

            self._lisp_decrement_()
            self._new_line_()
        self._lisp_decrement_()
        self._new_line_()
        self._lisp_decrement_()
        self._new_line_()

    def _output_port_(self, port):
        self._lisp_increment_()
        self._output_.write("port ")
        if len(port.pins) > 1:
            # TODO Clean up code in this if statement
            self._lisp_increment_()
            self._output_.write("array ")
            # self._lisp_increment_()
            # self._output_.write("rename ")
            self._output_name_of_object_(port)
            # self._output_.write(' "')
            # self._output_.write(port["EDIF.original_identifier"])
            # self._output_.write('"')
            # self._lisp_decrement_()
            self._output_.write(" ")
            self._output_.write(str(len(port.pins)))
            self._lisp_decrement_()
            self._output_.write(" ")
            self._lisp_increment_()
            self._output_.write("direction ")
            # self._output_.write(port.direction)
            self._output_.write(self._direction_to_string_(port.direction))  # str(port.direction))
            self._lisp_decrement_()
            self._lisp_decrement_()
            self._new_line_()
            return
        # self._output_.write(port.name)
        self._output_name_of_object_(port)
        self._lisp_increment_()
        self._output_.write("direction ")
        # self._output_.write(port.direction)
        self._output_.write(self._direction_to_string_(port.direction))  # str(port.direction))
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()

    def _direction_to_string_(self, direction):
        if direction == Port.Direction.INOUT:
            return "INOUT"
        elif direction == Port.Direction.IN:
            return "INPUT"
        elif direction == Port.Direction.OUT:
            return "OUTPUT"
        else:
            return "UNDEFINED"

    def _output_instance_(self, instance):
        self._lisp_increment_()
        self._output_.write("instance ")
        # print(vars(instance))
        # self._output_.write(self._get_edif_name_(instance))
        self._output_name_of_object_(instance)
        self._output_.write(" ")
        self._lisp_increment_()
        self._output_.write("viewref ")
        self._output_.write("netlist ")  # TODO this should be checked against some sort of metadata
        self._lisp_increment_()
        self._output_.write("cellref ")
        definition = instance.reference
        self._output_.write(self._get_edif_name_(definition))
        self._lisp_increment_()
        self._output_.write("libraryref ")
        self._output_.write(self._get_edif_name_(definition.library))
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()
        # print(vars(instance))
        if "EDIF.properties" in instance:
            properties = instance["EDIF.properties"]
            for prop in properties:
                self._output_property_(prop)
        self._lisp_decrement_()

    def _output_cable_(self, cable):
        self._lisp_increment_()
        self._output_.write("net ")
        if "EDIF.original_identifier" in cable:
            self._lisp_increment_()
            self._output_.write("rename ")
            self._output_.write(self._get_edif_name_(cable))
            self._output_.write(' "')
            self._output_.write(cable['EDIF.original_identifier'])
            self._output_.write('"')
            self._lisp_decrement_()
        else:
            self._output_.write(self._get_edif_name_(cable))
        self._output_.write(" ")
        self._lisp_increment_()
        self._output_.write("joined")
        self._new_line_()
        # for port in cable.getConnectionList(): #TODO fuction cable.getConnectionList() needs to be created
        for wire in cable.wires:
            for pin in wire.pins:
                # port = pin.port
                # print(type(pin))
                if isinstance(pin, OuterPin):
                    self._output_inner_pin_(pin)
                    pin = pin.inner_pin
                else:
                    self._output_port_ref_(pin.port, self._get_edif_name_(cable), pin)
        self._new_line_()
        self._lisp_decrement_()
        self._new_line_()
        self._lisp_decrement_()

    def _output_inner_pin_(self, pin):
        inner_pin = pin.inner_pin
        self._lisp_increment_()
        self._output_.write("portref ")
        if hasattr(pin.inner_pin.port, "is_array"):
            self._lisp_increment_()
            self._output_.write("member ")
            self._output_.write(self._get_edif_name_(inner_pin.port))
            for x in range(len(inner_pin.port.pins)):
                if inner_pin == inner_pin.port.pins[x]:
                    self._output_.write(" " + str(x))
                    self._lisp_decrement_()
        else:
            self._output_.write(self._get_edif_name_(inner_pin.port))
        self._output_.write(" ")
        self._lisp_increment_()
        self._output_.write("instanceref ")
        self._output_.write(pin.instance["EDIF.identifier"])
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()
        pass

    def _output_port_ref_(self, port_ref, cable_name, pin):
        self._lisp_increment_()
        self._output_.write("portref ")
        if hasattr(port_ref, "is_array"):
            for x in range(len(port_ref.pins)):
                # print(self.test)
                if port_ref.pins[x].wire is None:
                    # self.test += 1
                    # self._lisp_decrement_()
                    # print("test")
                    continue
                # if cable_name == port_ref.inner_pins[x].wire.cable["EDIF.identifier"]:
                if port_ref.pins[x] == pin:
                    break
            self._lisp_increment_()
            self._output_.write("member ")
            self._output_.write(self._get_edif_name_(port_ref))
            self._output_.write(" ")
            self._output_.write(str(x))
            self._lisp_decrement_()
        else:
            self._output_.write(self._get_edif_name_(port_ref))
        # self._output_.write(" ")
        # self._lisp_increment_()
        # self._output_.write("instanceref ")
        # self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()

    def _get_edif_name_(self, netlistObj):
        name = netlistObj["EDIF.identifier"]
        # print(vars(netlistObj))
        if not ("oldName" in netlistObj):
            return name
        else:
            oldName = netlistObj["oldName"]
            return "(rename " + name + ' "' + oldName + '")'

    def _lisp_increment_(self):
        self._output_.write("(")
        self._lisp_depth_ += 1

    def _new_line_(self):
        self._output_.write("\n")
        self._output_.write("  " * self._lisp_depth_)

    def _lisp_decrement_(self):
        self._output_.write(")")
        self._lisp_depth_ -= 1

    def _output_property_(self, prop):
        # TODO this only handles string properties for now
        self._lisp_increment_()
        self._output_.write("property ")
        if 'original_identifier' in prop:
            self._lisp_increment_()
            self._output_.write("rename ")
            self._output_.write(prop['identifier'])
            self._output_.write(' "')
            self._output_.write(prop['original_identifier'])
            self._output_.write('"')
            self._lisp_decrement_()
        else:
            self._output_.write(prop['identifier'])
        value = prop['value']
        self._output_.write(" ")
        self._lisp_increment_()
        if isinstance(value, str):
            self._output_.write('string "')
            self._output_.write(value)
            self._output_.write('"')
        elif isinstance(value, bool):
            self._output_.write('boolean ')
            self._lisp_increment_()
            self._output_.write(str(value))
            self._lisp_decrement_()
        else:
            self._output_.write('integer ')
            self._output_.write(str(value))
        self._lisp_decrement_()
        self._lisp_decrement_()
        self._new_line_()

    def _close_output_(self):
        self._output_.close()





from spydrnet.parsers.edif.tokenizer import EdifTokenizer
from spydrnet.parsers.edif.edif_tokens import *
from spydrnet.ir import *

from functools import reduce
import re


class EdifParser:
    def parse_construct(self, construct_parser):
        self.expect_begin_construct()
        result = construct_parser()
        self.expect_end_construct()
        return result

    @staticmethod
    def from_filename(filename):
        parser = EdifParser()
        parser.filename = filename
        return parser

    @staticmethod
    def from_file_handle(file_handle):
        parser = EdifParser()
        parser.file_handle = file_handle
        return parser

    def __init__(self):
        self.edif_identifier_namespace = dict() # class -> object -> subclass -> identifier -> object
        self.filename = None
        self.file_handle = None
        self.elements = list()
        self.tokenizer = None

    def parse(self):
        self.initialize_tokenizer()
        self.netlist = self.parse_construct(self.parse_edif)
        self.tokenizer.__del__()

    def initialize_tokenizer(self):
        if self.filename:
            self.tokenizer = EdifTokenizer.from_filename(self.filename)
        elif self.file_handle:
            self.tokenizer = EdifTokenizer.from_stream(self.file_handle)

    def parse_edif(self):
        environment = Netlist()
        self.append_new_element(environment)
        self.expect(EDIF)
        self.parse_nameDef()
        self.parse_header()
        self.parse_body()
        return self.pop_element()

    def parse_header(self):
        self.parse_construct(self.parse_edifVersion)
        self.parse_construct(self.parse_edifLevel)
        self.parse_construct(self.parse_keywordMap)

    def parse_edifVersion(self):
        self.expect(EDIF_VERSION)
        self.prefix_append('edifVersion')
        version_0 = self.parse_integerToken()
        version_1 = self.parse_integerToken()
        version_2 = self.parse_integerToken()
        self.set_attribute((version_0, version_1, version_2))
        self.prefix_pop()

    def parse_edifLevel(self):
        self.expect(EDIF_LEVEL)
        self.prefix_append('edifLevel')
        level = self.parse_integerToken()
        if level != 0:
            self.set_attribute(level)
        self.prefix_pop()

    def parse_keywordMap(self):
        self.expect(KEYWORD_MAP)
        self.prefix_append('keywordMap')
        self.parse_construct(self.parse_keywordLevel)

        while self.begin_construct():
            self.parse_construct(self.parse_comment)
        self.prefix_pop()

    def parse_keywordLevel(self):
        self.expect(KEYWORD_LEVEL)
        self.prefix_append('keywordLevel')
        level = self.parse_integerToken()
        if level != 0:
            self.set_attribute(level)
        self.prefix_pop()

    def parse_body(self):
        has_status = False
        while self.begin_construct():
            if self.construct_is(STATUS):
                has_status = self.check_for_multiples(STATUS, has_status)
                self.parse_status()

            elif self.construct_is(LIBRARY):
                library = self.parse_library()
                environment = self.elements[-1]
                environment.add_library(library)
                self.add_element_to_namespace(library)
            elif self.construct_is(EXTERNAL):
                library = self.parse_external()
                environment = self.elements[-1]
                environment.add_library(library)
                self.add_element_to_namespace(library)
            elif self.construct_is(DESIGN):
                self.parse_design()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect("|".join([STATUS, EXTERNAL, LIBRARY, DESIGN, COMMENT, USER_DATA]))
            self.expect_end_construct()

    def parse_status(self):
        self.expect(STATUS)
        self.prefix_append('status')
        while self.begin_construct():
            if self.construct_is(WRITTEN):
                self.parse_written()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect("|".join([WRITTEN | COMMENT | USER_DATA]))
            self.expect_end_construct()
        self.prefix_pop()

    def parse_written(self):
        self.expect(WRITTEN)
        self.prefix_append('written')
        self.parse_construct(self.parse_timeStamp)

        has_author = False
        has_program = False
        has_dataOrigin = False
        while self.begin_construct():
            if self.construct_is(AUTHOR):
                has_author = self.check_for_multiples(AUTHOR, has_author)
                self.parse_author()

            elif self.construct_is(PROGRAM):
                has_program = self.check_for_multiples(PROGRAM, has_program)
                self.parse_program()

            elif self.construct_is(DATA_ORIGIN):
                has_dataOrigin = self.check_for_multiples(DATA_ORIGIN, has_dataOrigin)
                self.parse_dataOrigin()
            elif self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(METAX):
                self.parse_metax()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect("|".join([AUTHOR, PROGRAM, DATA_ORIGIN, PROPERTY, METAX, COMMENT, USER_DATA]))
            self.expect_end_construct()
        self.prefix_pop()

    def parse_timeStamp(self):
        self.expect(TIME_STAMP)
        self.prefix_append('timeStamp')
        year = self.parse_integerToken()
        month = self.parse_integerToken()
        day = self.parse_integerToken()
        hour = self.parse_integerToken()
        minute = self.parse_integerToken()
        second = self.parse_integerToken()
        self.set_attribute((year, month, day, hour, minute, second))
        self.prefix_pop()

    def parse_author(self):
        self.expect(AUTHOR)
        self.prefix_append('author')
        author = self.parse_stringToken()
        self.set_attribute(author)
        self.prefix_pop()

    def parse_program(self):
        self.expect(PROGRAM)
        self.prefix_append('program')
        program = self.parse_stringToken()
        self.set_attribute(program)

        if self.begin_construct():
            self.expect(VERSION)
            self.prefix_append('version')
            version = self.parse_stringToken()
            self.set_attribute(version)
            self.prefix_pop()
            self.expect_end_construct()
        self.prefix_pop()

    def parse_library(self):
        self.expect(LIBRARY)
        return self.parse_library_like_element()

    def parse_external(self):
        self.expect(EXTERNAL)
        return self.parse_library_like_element(is_external=True)

    def parse_library_like_element(self, is_external=False):
        library = Library()
        if is_external:
            library['EDIF.external'] = True
        self.append_new_element(library)

        self.parse_nameDef()
        self.parse_construct(self.parse_edifLevel)
        self.parse_construct(self.parse_technology)

        has_status = False
        while self.begin_construct():
            if self.construct_is(STATUS):
                has_status = self.check_for_multiples(STATUS, has_status)
                self.parse_status()

            elif self.construct_is(CELL):
                definition = self.parse_cell()
                library = self.elements[-1]
                library.add_definition(definition)
                self.add_element_to_namespace(definition)
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect("|".join([STATUS, CELL, COMMENT, USER_DATA]))
            self.expect_end_construct()

        return self.pop_element()

    def parse_technology(self):
        self.expect(TECHNOLOGY)
        self.parse_construct(self.parse_numberDefinition)

    def parse_numberDefinition(self):
        self.expect(NUMBER_DEFINITION)
        self.skip_until_next_construct()

    def parse_cell(self):
        definition = Definition()
        self.append_new_element(definition)

        self.expect(CELL)
        self.parse_nameDef()
        self.parse_construct(self.parse_cellType)

        has_status = False
        has_viewMap = False
        while self.begin_construct():
            if self.construct_is(STATUS):
                has_status = self.check_for_multiples(STATUS, has_status)
                self.parse_status()

            elif self.construct_is(VIEW):
                self.parse_view()
            elif self.construct_is(VIEW_MAP):
                has_viewMap = self.check_for_multiples(VIEW_MAP, has_viewMap)
                self.parse_viewMap()

            elif self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([STATUS, VIEW, VIEW_MAP, PROPERTY, COMMENT, USER_DATA]))
            self.expect_end_construct()

        return self.pop_element()

    def parse_cellType(self):
        self.expect(CELL_TYPE)
        self.prefix_append("cellType")
        if self.construct_is(GENERIC) \
                or self.construct_is(TIE) \
                or self.construct_is(RIPPER):
            if not self.tokenizer.token_equals(GENERIC):
                self.set_attribute(self.tokenizer.token)
        else:
            self.expect("|".join([GENERIC, TIE, RIPPER]))
        self.tokenizer.next()
        self.prefix_pop()

    def parse_view(self):
        self.expect(VIEW)
        self.prefix_append('view')
        self.parse_nameDef()
        self.parse_construct(self.parse_viewType)
        self.parse_construct(self.parse_interface)

        has_status = False
        has_contents = False
        while self.begin_construct():
            if self.construct_is(STATUS):
                has_status = self.check_for_multiples(STATUS, has_status)
                self.parse_status()

            elif self.construct_is(CONTENTS):
                has_contents = self.check_for_multiples(STATUS, has_contents)
                self.parse_contents()

            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([STATUS, CONTENTS, COMMENT, PROPERTY, USER_DATA]))
            self.expect_end_construct()
        self.prefix_pop()

    def parse_viewType(self):
        self.prefix_append('viewType')
        self.expect(VIEW_TYPE)
        if self.construct_is(BEHAVIOR) \
                or self.construct_is(DOCUMENT) \
                or self.construct_is(GRAPHIC) \
                or self.construct_is(LOGICMODEL) \
                or self.construct_is(MASKLAYOUT) \
                or self.construct_is(NETLIST) \
                or self.construct_is(PCBLAYOUT) \
                or self.construct_is(SCHEMATIC) \
                or self.construct_is(STRANGER) \
                or self.construct_is(SYMBOLIC):
            if not self.tokenizer.token_equals(NETLIST):
                self.set_attribute(self.tokenizer.token)
        else:
            self.expect("|".join(
                [BEHAVIOR, DOCUMENT, GRAPHIC, LOGICMODEL, MASKLAYOUT, NETLIST, PCBLAYOUT, SCHEMATIC, STRANGER,
                 SYMBOLIC]))
        self.tokenizer.next()
        self.prefix_pop()

    def parse_interface(self):
        self.expect(INTERFACE)
        has_designator = False
        while self.begin_construct():
            if self.construct_is(PORT):
                port = self.parse_port()
                cell = self.elements[-1]
                cell.add_port(port)
                self.add_element_to_namespace(port)
            elif self.construct_is(PORT_BUNDLE):
                raise NotImplementedError()
            elif self.construct_is(SYMBOL):
                raise NotImplementedError()
            elif self.construct_is(PROTECTION_FRAME):
                raise NotImplementedError()
            elif self.construct_is(ARRAY_RELATED_INFO):
                raise NotImplementedError()
            elif self.construct_is(PARAMETER):
                raise NotImplementedError()
            elif self.construct_is(JOINED):
                raise NotImplementedError()
            elif self.construct_is(MUST_JOIN):
                raise NotImplementedError()
            elif self.construct_is(WEAK_JOINED):
                raise NotImplementedError()
            elif self.construct_is(PERMUTABLE):
                raise NotImplementedError()
            elif self.construct_is(TIMING):
                raise NotImplementedError()
            elif self.construct_is(SIMULATE):
                raise NotImplementedError()
            elif self.construct_is(DESIGNATOR):
                has_designator = self.check_for_multiples(DESIGNATOR, has_designator)
                self.parse_designator()
            elif self.construct_is(WEAK_JOINED):
                raise NotImplementedError()

            elif self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([PORT, PROPERTY, COMMENT, USER_DATA]))
            self.expect_end_construct()

    def parse_designator(self):
        self.expect(DESIGNATOR)
        self.skip_until_next_construct()

    def parse_port(self):
        self.append_new_element(Port())
        self.expect(PORT)
        if self.begin_construct():
            if self.construct_is(RENAME):
                self.parse_rename()
                port = self.elements[-1]
                port.create_pins(1)

            elif self.construct_is(ARRAY):
                dimension_sizes = self.parse_array()
                pin_count = reduce((lambda x, y: x * y), dimension_sizes)
                port = self.elements[-1]
                port.create_pins(pin_count)
                port.is_array = True
                if 'EDIF.original_identifier' in port:
                    # TODO: what about multi-dimensional ports, non-downto ports, and when non-square brackets are used <0:17><31:0>
                    original_identifier = port['EDIF.original_identifier']
                    match = re.match(r".*\[(\d+):(\d+)\]", original_identifier)
                    if match:
                        left_index = int(match.group(1))
                        right_index = int(match.group(2))
                        port.lower_index = min(right_index, left_index)

            else:
                self.expect('|'.join([RENAME, ARRAY]))
            self.expect_end_construct()
        else:
            self.parse_nameDef()
            port = self.elements[-1]
            port.create_pins(1)
            # TODO: what about single pin array ports with a non_zero starting index.

        has_direction = False
        while self.begin_construct():
            if self.construct_is(DIRECTION):
                has_direction = self.check_for_multiples(DIRECTION, has_direction)
                direction = self.parse_direction()
                port = self.elements[-1]
                port.direction = direction

            elif self.construct_is(UNUSED):
                raise NotImplementedError()
            elif self.construct_is(DESIGNATOR):
                raise NotImplementedError()
            elif self.construct_is(DC_FANIN_LOAD):
                raise NotImplementedError()
            elif self.construct_is(DC_FANOUT_LOAD):
                raise NotImplementedError()
            elif self.construct_is(DC_MAX_FANIN):
                raise NotImplementedError()
            elif self.construct_is(DC_MAX_FANOUT):
                raise NotImplementedError()
            elif self.construct_is(AC_LOAD):
                raise NotImplementedError()
            elif self.construct_is(PORT_DELAY):
                raise NotImplementedError()

            elif self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([DIRECTION, PROPERTY, COMMENT, USER_DATA]))
            self.expect_end_construct()
        return self.elements.pop()

    def parse_array(self):
        self.expect(ARRAY)
        self.parse_nameDef()
        dimension_sizes = [self.parse_integerToken()]
        while self.tokenizer.is_valid_identifier():
            dimension_sizes.append(self.parse_integerToken())
        return dimension_sizes

    def parse_direction(self):
        self.expect(DIRECTION)
        direction = Port.Direction.UNDEFINED
        if self.construct_is(INOUT):
            direction = Port.Direction.INOUT
        elif self.construct_is(INPUT):
            direction = Port.Direction.IN
        elif self.construct_is(OUTPUT):
            direction = Port.Direction.OUT
        else:
            self.expect('|'.join([INOUT, INPUT, OUTPUT]))
        self.tokenizer.next()
        return direction

    def parse_contents(self):
        self.expect(CONTENTS)
        while self.begin_construct():
            if self.construct_is(INSTANCE):
                instance = self.parse_instance()
                definition = self.elements[-1]
                definition.add_child(instance)
                self.add_element_to_namespace(instance)
            elif self.construct_is(NET):
                cable = self.parse_net()
                definition = self.elements[-1]
                try:
                    definition.add_cable(cable)
                except KeyError:
                    print()
                    raise KeyError("Error on line",  self.tokenizer.line_number)

            elif self.construct_is(OFF_PAGE_CONNECTOR):
                raise NotImplementedError()
            elif self.construct_is(FIGURE):
                raise NotImplementedError()
            elif self.construct_is(SECTION):
                raise NotImplementedError()
            elif self.construct_is(NET_BUNDLE):
                raise NotImplementedError()
            elif self.construct_is(PAGE):
                raise NotImplementedError()
            elif self.construct_is(COMMENT_GRAPHICS):
                raise NotImplementedError()
            elif self.construct_is(PORT_IMPLEMENTATION):
                raise NotImplementedError()
            elif self.construct_is(TIMING):
                raise NotImplementedError()
            elif self.construct_is(SIMULATE):
                raise NotImplementedError()
            elif self.construct_is(WHEN):
                raise NotImplementedError()
            elif self.construct_is(FOLLOW):
                raise NotImplementedError()
            elif self.construct_is(LOGIC_PORT):
                raise NotImplementedError()
            elif self.construct_is(BOUNDING_BOX):
                raise NotImplementedError()
            elif self.construct_is(TIMING):
                raise NotImplementedError()

            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([INSTANCE, NET, COMMENT, USER_DATA]))
            self.expect_end_construct()

    def parse_instance(self):
        self.append_new_element(Instance())
        self.expect(INSTANCE)
        self.parse_nameDef()
        if self.begin_construct():
            if self.construct_is(VIEW_REF):
                definition = self.parse_viewRef()
                instance = self.elements[-1]
                instance.reference = definition

            elif self.construct_is(VIEW_LIST):
                raise NotImplementedError()
            else:
                self.expect(VIEW_REF)
            self.expect_end_construct()

        while self.begin_construct():
            if self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([PROPERTY, COMMENT, USER_DATA]))
            self.expect_end_construct()
        return self.pop_element()

    def parse_viewRef(self):
        self.prefix_append('viewRef')
        self.expect(VIEW_REF)
        self.parse_nameRef()
        view_identifier = self.elements[-1].pop('EDIF.viewRef.identifier')
        definition = self.elements[-2]
        if self.begin_construct():
            definition = self.parse_cellRef()
            self.expect_end_construct()
        if definition['EDIF.view.identifier'].lower() != view_identifier.lower():
            raise RuntimeError("Parser error, non-existant view referenced on line {}, revieved {} expected {}".format(
                self.tokenizer.line_number, view_identifier, definition['EDIF.view.identifier']
            ))
        self.prefix_pop()
        return definition

    def parse_cellRef(self):
        self.prefix_append('cellRef')
        self.expect(CELL_REF)
        self.parse_nameRef()
        definition_identifer = self.elements[-1].pop('EDIF.viewRef.cellRef.identifier')
        library = self.elements[-3]
        if self.begin_construct():
            library = self.parse_libraryRef()
            self.expect_end_construct()
        definition = self.get_element_by_identifier(library, Definition, definition_identifer)
        self.prefix_pop()
        return definition

    def parse_libraryRef(self):
        self.prefix_append('libraryRef')
        self.expect(LIBRARY_REF)
        self.parse_nameRef()
        library_identifier = self.elements[-1].pop('EDIF.viewRef.cellRef.libraryRef.identifier')
        environment = self.elements[-4]
        library = self.elements[-3]
        if library['EDIF.identifier'].lower() != library_identifier.lower():
            library = self.get_element_by_identifier(environment, Library, library_identifier)
        self.prefix_pop()
        return library

    def parse_net(self):
        self.append_new_element(Cable())
        self.expect(NET)
        self.parse_nameDef()
        self.elements[-1].is_scalar = True
        self.elements[-1].create_wires(1)  # EDIF nets are single wire cables.
        self.parse_construct(self.parse_joined)

        while self.begin_construct():
            if self.construct_is(PROPERTY):
                self.parse_property()
            elif self.construct_is(COMMENT):
                self.parse_comment()
            elif self.construct_is(USER_DATA):
                self.parse_userData()
            else:
                self.expect('|'.join([PROPERTY, COMMENT, USER_DATA]))
            self.expect_end_construct()
        return self.pop_element()

    def parse_joined(self):
        self.expect(JOINED)
        while self.begin_construct():
            if self.construct_is(PORT_REF):
                pin = self.parse_portRef()
                wire = self.elements[-1].wires[0]
                wire.connect_pin(pin)
            elif self.construct_is(PORT_LIST):
                raise NotImplementedError()
            elif self.construct_is(GLOBAL_PORT_REF):
                raise NotImplementedError()
            else:
                self.expect(PORT_REF)
            self.expect_end_construct()

    def parse_portRef(self):
        self.prefix_append('portRef')
        self.expect(PORT_REF)
        index = 0
        instance_or_definition = self.elements[-2]
        if self.begin_construct():
            indicies = self.parse_member()
            assert (len(indicies) == 1)
            index = indicies[0]
            self.expect_end_construct()
        else:
            self.parse_nameRef()

        while self.begin_construct():
            if self.construct_is(PORT_REF):
                raise NotImplementedError()
            elif self.construct_is(INSTANCE_REF):
                instance_or_definition = self.parse_instanceRef()
            elif self.construct_is(VIEW_REF):
                raise NotImplementedError()
            self.expect_end_construct()
        port_identifier = self.elements[-1].pop('EDIF.portRef.identifier')
        if isinstance(instance_or_definition, Instance):
            definition = instance_or_definition.reference
            port = self.get_element_by_identifier(definition, Port, port_identifier)
            inner_pin = port.pins[index]
            pin = instance_or_definition.pins[inner_pin]
        else:
            port = self.get_element_by_identifier(instance_or_definition, Port, port_identifier)
            pin = port.pins[index]
        self.prefix_pop()
        return pin

    def parse_instanceRef(self):
        self.prefix_append('instanceRef')
        definition = self.elements[-2]
        self.expect(INSTANCE_REF)
        if self.begin_construct():
            self.parse_member()
            raise NotImplementedError()
            self.expect_end_construct()
        else:
            self.parse_nameRef()
        instance_identifier = self.elements[-1].pop('EDIF.portRef.instanceRef.identifier')
        instance = self.get_element_by_identifier(definition, Instance, instance_identifier)
        self.prefix_pop()
        return instance

    def parse_member(self):
        self.expect(MEMBER)
        self.parse_nameDef()
        indicies = [self.parse_integerToken()]
        while self.not_end_construct():
            indicies.append(self.parse_integerToken())
            self.expect_end_construct()
        return indicies

    def parse_viewMap(self):
        self.expect(VIEW_MAP)
        raise NotImplementedError()

    def parse_design(self):
        self.expect(DESIGN)
        # self.tokenizer.next()
        instance = Instance()
        instance['metadata_prefix'] = list()
        self.elements.append(instance)
        if self.begin_construct():
            instance['metadata_prefix'] = ['EDIF']
            self.parse_rename()
            instance['metadata_prefix'] = []
            self.tokenizer.next()
        else:
            instance['EDIF.identifier'] = self.tokenizer.next()
        self.tokenizer.next()
        self.tokenizer.next()
        definition_name = self.tokenizer.next()
        self.tokenizer.next()
        self.tokenizer.next()
        library_name = self.tokenizer.next()
        for library in self.elements[0].libraries:
            if library['EDIF.identifier'] == library_name:
                break
        for definition in library.definitions:
            if definition['EDIF.identifier'] == definition_name:
                break
        instance.reference = definition
        self.elements.pop()
        self.elements[0].top_instance = instance
        self.skip_until_next_construct()

    def parse_dataOrigin(self):
        self.expect(DATA_ORIGIN)
        raise NotImplementedError()

    def parse_userData(self):
        self.expect(USER_DATA)
        raise NotImplementedError()

    def parse_comment(self):
        self.prefix_append('comments')
        self.expect(COMMENT)
        comment = list()
        while self.not_end_construct():
            comment.append(self.parse_stringToken())
        comment = (*comment,)
        self.append_attribute(comment)
        self.prefix_pop()

    def parse_property(self):
        self.prefix_append('properties')
        self.expect(PROPERTY)
        self.parse_property_like_element()

    def parse_metax(self):
        self.prefix_append('metaxes')
        self.expect(METAX)
        self.parse_property_like_element()

    def parse_property_like_element(self):
        self.parse_nameDef()

        property_ = dict()
        identifier = self.elements[-1].pop('.'.join([*self.elements[-1]['metadata_prefix'], 'identifier']))
        property_['identifier'] = identifier

        original_identifier_prefix = '.'.join([*self.elements[-1]['metadata_prefix'], 'original_identifier'])
        if original_identifier_prefix in self.elements[-1]:
            original_identifier = self.elements[-1].pop(original_identifier_prefix)
            property_['original_identifier'] = original_identifier

        value = self.parse_construct(self.parse_typedValue)
        property_['value'] = value

        self.append_attribute(property_)

        has_owner = False
        has_unit = False
        while self.begin_construct():
            if self.construct_is(OWNER):
                has_owner = self.check_for_multiples(OWNER, has_owner)
                self.parse_owner()

            elif self.construct_is(UNIT):
                has_unit = self.check_for_multiples(UNIT, has_unit)
                raise NotImplementedError()  # self.parse_unit()

            elif self.construct_is(PROPERTY):
                raise NotImplementedError()  # self.parse_property()
            elif self.construct_is(COMMENT):
                raise NotImplementedError()  # self.parse_comment()
            self.expect_end_construct()
        self.prefix_pop()

    def parse_typedValue(self):
        if self.construct_is(BOOLEAN):
            return self.parse_boolean()
        elif self.construct_is(INTEGER):
            return self.parse_integer()
        elif self.construct_is(MI_NO_MAX):
            raise NotImplementedError()
        elif self.construct_is(NUMBER):
            return self.parse_number()
        elif self.construct_is(POINT):
            raise NotImplementedError()
        elif self.construct_is(STRING):
            return self.parse_string()
        else:
            return self.expect('|'.join([BOOLEAN, INTEGER, NUMBER, STRING]))

    def parse_boolean(self):
        self.expect(BOOLEAN)
        self.expect_begin_construct()
        self.tokenizer.next()
        if self.tokenizer.token_equals(TRUE):
            result = True
        elif self.tokenizer.token_equals(FALSE):
            result = False
        else:
            self.expect('|'.join([TRUE, FALSE]))
        self.expect_end_construct()
        return result

    def parse_integer(self):
        self.expect(INTEGER)
        return self.parse_integerToken()

    def parse_number(self):
        self.expect(NUMBER)
        if self.begin_construct():
            result = self.parse_construct(self.parse_e)
            self.expect_end_construct()
        else:
            result = self.parse_integerToken()
        return result

    def parse_e(self):
        self.expect(E)
        mantissa = self.parse_integerToken()
        exponent = self.parse_integerToken()
        result = mantissa * 10.0 ** exponent
        return result

    def parse_string(self):
        self.expect(STRING)
        return self.parse_stringToken()

    def parse_owner(self):
        self.expect(OWNER)
        self.parse_stringToken()

    def parse_unit(self):
        raise NotImplementedError()

    def parse_nameRef(self):
        self.prefix_append('identifier')
        self.set_attribute(self.parse_identifier())
        self.prefix_pop()

    def parse_nameDef(self):
        if self.begin_construct():
            self.parse_rename()
            self.expect_end_construct()
        else:
            self.prefix_append('identifier')
            self.set_attribute(self.parse_identifier())
            self.prefix_pop()

    def parse_rename(self):
        self.expect(RENAME)

        self.prefix_append('identifier')
        self.set_attribute(self.parse_identifier())
        self.prefix_pop()

        self.prefix_append('original_identifier')
        self.set_attribute(self.parse_stringToken())
        self.prefix_pop()

    def parse_identifier(self):
        self.tokenizer.next()
        self.tokenizer.expect_valid_identifier()
        return self.tokenizer.token

    def parse_stringToken(self):
        self.tokenizer.next()
        self.tokenizer.expect_valid_stringToken()
        return self.tokenizer.token[1:-1]

    def parse_integerToken(self):
        self.tokenizer.next()
        self.tokenizer.expect_valid_integerToken()
        return int(self.tokenizer.token)

    def append_new_element(self, element):
        element['metadata_prefix'] = ['EDIF']
        self.elements.append(element)

    def pop_element(self):
        element = self.elements.pop()
        del element['metadata_prefix']
        return element

    def prefix_append(self, value):
        element = self.elements[-1]
        element['metadata_prefix'].append(value)

    def prefix_pop(self):
        return self.elements[-1]['metadata_prefix'].pop()

    def set_attribute(self, value):
        element = self.elements[-1]
        element['.'.join(element['metadata_prefix'])] = value

    def append_attribute(self, attribute):
        element = self.elements[-1]
        key = '.'.join(element['metadata_prefix'])
        if key not in element:
            element[key] = list()
        element[key].append(attribute)

    def skip_until_next_construct(self):
        count = 0
        while count > 0 or not self.tokenizer.peek_equals(RIGHT_PAREN):
            if self.tokenizer.peek_equals(LEFT_PAREN):
                count += 1
            elif self.tokenizer.peek_equals(RIGHT_PAREN):
                count -= 1
            self.tokenizer.next()

    def check_for_multiples(self, token, already_contains):
        if already_contains:
            raise RuntimeError(
                "Parse error: Multiple occurances of {}, near line {}".format(token, self.tokenizer.line_number))
        return True

    def expect_begin_construct(self):
        self.tokenizer.next()
        if LEFT_PAREN != self.tokenizer.token:
            self.tokenizer.expect(LEFT_PAREN)

    def not_end_construct(self):
        if RIGHT_PAREN != self.tokenizer.peek():
            return True
        return False

    def expect_end_construct(self):
        self.tokenizer.next()
        if RIGHT_PAREN != self.tokenizer.token:
            self.tokenizer.expect(RIGHT_PAREN)

    def expect(self, token):
        self.tokenizer.next()
        self.tokenizer.expect(token)

    def begin_construct(self):
        if LEFT_PAREN == self.tokenizer.peek():
            self.tokenizer.next()
            return True
        return False

    def construct_is(self, token):
        if self.tokenizer.peek_equals(token):
            return True
        return False

    get_parent = {Library: lambda x: x.netlist,
                  Definition: lambda x: x.library,
                  Port: lambda x: x.definition,
                  Cable: lambda x: x.definition,
                  Instance: lambda x: x.parent}

    def add_element_to_namespace(self, element):
        parent = self.get_parent[type(element)](element)
        if parent in self.edif_identifier_namespace:
            parent_namespace = self.edif_identifier_namespace[parent]
        else:
            parent_namespace = dict()
            self.edif_identifier_namespace[parent] = parent_namespace
        element_type = type(element)
        if element_type in parent_namespace:
            element_namespace = parent_namespace[element_type]
        else:
            element_namespace = dict()
            parent_namespace[element_type] = element_namespace
        identifier = element['EDIF.identifier']
        identifier_lower = identifier.lower()
        assert identifier_lower not in element_namespace, "Identifier collision"
        element_namespace[identifier_lower] = element

    def get_element_by_identifier(self, parent, element_type, identifier):
        assert parent in self.edif_identifier_namespace, "No name space associated with parent element"
        parent_namespace = self.edif_identifier_namespace[parent]
        assert element_type in parent_namespace, "No elements in parent namespace of element type"
        element_namespace = parent_namespace[element_type]
        identifier_lower = identifier.lower()
        assert identifier_lower in element_namespace, "No element by name of identifier"
        return element_namespace[identifier_lower]



if __name__ == "__main__":
    # filename = r"C:\Users\keller\workplace\SpyDrNet\data\large_edif\osfbm.edf"
    filename = r"C:\Users\akeller9\workspace\SpyDrNet\data\large_edif\osfbm.edf"
    parser = EdifParser.from_filename(filename)
    import cProfile

    cProfile.run('parser.parse()')


#COPYRIGHT BYUCCL 2020

from spydrnet.callback.callback_listener import CallbackListener
from spydrnet.ir import Library, Definition, Instance, Port, Cable

class DataManager(CallbackListener):
    '''
    this class is a plugin using the callback listener to manage a database for name management.
    
    This class implements and registers all listeners and presents to the user
    the ability to do quick lookups based on name.
    '''

    ######################################################
    #User access functions to use datamanager
    ######################################################

    
    ######################################################
    #Override funcitons to register listeners
    ######################################################

    def __init__(self):
        # dictionary for each namespace
        # netlist simple dictionary
        self.net_dict = dict()
        # library dictionary from library to dictionary from name to value
        self.lib_dict = dict()
        # definition dictionary from defintion to dictionary from name to value
        self.defp_dict = dict()
        self.defc_dict = dict()
        self.defi_dict = dict()
        # call this to register the listeners
        super().__init__() #calls register_all_listeners
    

    def register_all_listeners(self):
        '''override the function that registers what listeners are available.'''
        self.register_definition_add_port()
        self.register_definition_remove_port()
        self.register_definition_add_child()
        self.register_definition_remove_child()
        self.register_definition_add_cable()
        self.register_definition_remove_cable()
        self.register_library_add_definition()
        self.register_library_remove_definition()
        self.register_netlist_add_library()
        self.register_netlist_remove_library()
        self.register_dictionary_delete()
        self.register_dictionary_set()
        self.register_dictionary_pop()


    ######################################################
    #Override functions that listen
    ######################################################

    def definition_add_port(self, definition, port):
        self.add_to_dict(self.defp_dict, definition, port)

    def definition_remove_port(self, definition, port):
        self.remove_from_dict(self.defp_dict, definition, port)

    def definition_add_child(self, definition, child):
        self.add_to_dict(self.defi_dict, definition, child)

    def definition_remove_child(self, definition, child):
        self.remove_from_dict(self.defi_dict, definition, child)

    def definition_add_cable(self, definition, cable):
        self.add_to_dict(self.defc_dict, definition, cable)

    def definition_remove_cable(self, definition, cable):
        self.remove_from_dict(self.defc_dict, definition, cable)

    def library_add_definition(self, library, definition):
        self.add_to_dict(self.lib_dict, library, definition)

    def library_remove_definition(self, library, definition):
        self.remove_from_dict(self.lib_dict, library, definition)

    def dictionary_delete(self, element, key):
        if key != 'EDIF.identifier':
            return
        self.key_remover(element,key)

    def dictionary_set(self, element, key, value):
        if key != 'EDIF.identifier':
            return
        if "EDIF.identifier" in element:
            #this is a rename we must remove first.
            self.dictionary_delete(element, "EDIF.identifier")
        if isinstance(element, Cable):
            if element.definition is None:
                return
            self.dict_set(self.defc_dict, element.definition, element, value)
        if isinstance(element, Definition):
            if element.library is None:
                return
            self.dict_set(self.lib_dict, element.library, element, value)
        if isinstance(element, Library):
            if element.netlist is None:
                return
            self.net_dict_set(self.net_dict, element.netlist, element, value)
        if isinstance(element, Port):
            if element.definition is None:
                return
            self.dict_set(self.defp_dict, element.definition, element, value)
        if isinstance(element, Instance):
            if element.parent is None:
                return
            self.dict_set(self.defi_dict, element.parent, element, value)

    def dictionary_pop(self, element, item):
        key = item
        if key != 'EDIF.identifier':
            return
        self.key_remover(element,key)

    def netlist_add_library(self, netlist, library):
        #add the library["EDIF.identifier"] to the data structure for netlists
        if "EDIF.identifier" not in library:
            return
        self.net_dict_set(self.net_dict, netlist, library, library["EDIF.identifier"])

    def netlist_remove_library(self, netlist, library):
        if "EDIF.identifier" not in library:
            return
        #remove the library["EDIF.identifier"] from the data structure for netlists
        assert library["EDIF.identifier"] in self.net_dict and self.net_dict[library["EDIF.identifier"]] == library, "Library not present in given object netlist"
        self.net_dict.pop(library["EDIF.identifier"])


    #################################################################################
    # Helper functions
    #################################################################################

    def add_to_dict(self, container, parent, child):
        if "EDIF.identifier" not in child:
            return
        self.dict_set(container, parent, child, child["EDIF.identifier"])

    def key_remover(self, element, key):
        if isinstance(element, Cable):
            self.remove_from_dict(self.defc_dict, element.definition, element)
        if isinstance(element, Definition):
            self.remove_from_dict(self.lib_dict, element.library, element)
        if isinstance(element, Library):
            self.netlist_remove_library(element.netlist, element)
        if isinstance(element, Port):
            self.remove_from_dict(self.defp_dict, element.definition, element)
        if isinstance(element, Instance):
            self.remove_from_dict(self.defi_dict, element.parent, element)

    def remove_from_dict(self, container, parent, child):
        if "EDIF.identifier" not in child:
            return
        assert parent in container and child["EDIF.identifier"] in container[parent], "Child object " + child.__class__.__name__ + " not present in given object's namespace " + parent.__class__.__name__
        container[parent].pop(child["EDIF.identifier"])
    

    def dict_set(self, container, parent, child, value):
        if parent in container:
            if value in container[parent]:
               raise ValueError("Edif namespace violation while adding " + child.__class__.__name__ + " to " + parent.__class__.__name__)
            else:
                container[parent][value] = child
        else:
            container[parent] = dict()
            container[parent][value] = child

    def net_dict_set(self, container, parent, child, value):
        if(value in self.net_dict):
            raise ValueError("Edif namespace violation while adding Library to Netlist")
        else:
            self.net_dict[value] = child


import os

def compose(filename, netlist):
    extension = os.path.splitext(filename)[1]
    extension_lower = extension.lower()
    if extension_lower in {".edf", ".edif"}:
        from spydrnet.composers.edif.composer import ComposeEdif
        composer = ComposeEdif()
        composer.run(netlist, filename)
    else:
        raise RuntimeError("Extension {} not recognized.".format(extension))



from spydrnet.ir.pin import Pin


class OuterPin(Pin):
    """
    Pins that correspond to instances. These pins can be thought of as on the outside of an instance. There can be many
    outer pins for each inner pin
    """
    __slots__ = ['_instance', '_inner_pin']

    @staticmethod
    def from_instance_and_inner_pin(instance, inner_pin):
        '''create an outer pin associated with a given inner_pin and instance object.

        parameters
        ----------

        instance - (Instance) the instance to associate with this pin

        inner_pin - (InnerPin) the inner pin with which to associate this outer pin'''
        return OuterPin(instance, inner_pin)

    def __init__(self, instance=None, inner_pin=None):
        '''create an OuterPin.

        parameters
        ----------

        instance - (Instance) the instance with which to associate this outper pin.

        inner_pin - (InnerPin) a definition's inner pin to be associated with this instance outer pin.'''
        super().__init__()
        self._instance = instance
        self._inner_pin = inner_pin

    @property
    def instance(self):
        '''Return the instance with which this pin is associated'''
        return self._instance

    @property
    def inner_pin(self):
        '''get the inner pin associated with this outer pin'''
        return self._inner_pin

    def __eq__(self, other):
        if isinstance(other, OuterPin):
            return self._instance == other._instance and self._inner_pin == other._inner_pin
        return False

    def __hash__(self):
        return hash((self._instance, self._inner_pin))


import unittest

import spydrnet as sdn
from spydrnet.ir.element import Element


class TestLibrary(unittest.TestCase):
    def setUp(self):
        self.library = sdn.Library()

    def test_constructor(self):
        self.assertIsInstance(self.library, Element, "Netlist is not an element.")
        self.assertTrue(self.library, "Constructor return None type or empty collection")
        library2 = sdn.Netlist()
        self.assertNotEqual(self.library, library2, "Unique objects are considered equal.")

    def test_definitions_set(self):
        definition1 = self.library.create_definition()
        definition2 = self.library.create_definition()
        definitions = [definition1, definition2]
        self.assertEqual(self.library.definitions, definitions)
        self.library.definitions = reversed(definitions)
        self.assertEqual(self.library.definitions, list(reversed(definitions)))

    def test_create_definition(self):
        definition = self.library.create_definition()
        self.assertTrue(definition in self.library.definitions)
        self.assertEqual(definition.library, self.library)

    def test_add_definition(self):
        definition = sdn.Definition()
        self.library.add_definition(definition)
        self.assertTrue(definition in self.library.definitions)
        self.assertEqual(definition.library, self.library)
        self.assertEqual(list(self.library.definitions).count(definition), 1)

        definition = sdn.Definition()
        self.library.add_definition(definition, position=0)
        self.assertTrue(definition in self.library.definitions)
        self.assertEqual(definition.library, self.library)
        self.assertEqual(list(self.library.definitions).count(definition), 1)

    def test_remove_definition(self):
        definition = self.library.create_definition()
        self.library.remove_definition(definition)
        self.assertFalse(definition in self.library)
        self.assertIsNone(definition.library)

    @unittest.expectedFailure
    def test_remove_definitions_from_outside_library(self):
        definition = sdn.Definition()
        self.library.remove_definitions_from([definition])

    def test_remove_definitions_from(self):
        definition_included = self.library.create_definition()
        definition = self.library.create_definition()
        self.library.remove_definitions_from({definition})
        self.assertFalse(definition in self.library.definitions)
        self.assertIsNone(definition.library)
        self.assertTrue(definition_included in self.library.definitions)
        self.assertEqual(definition_included.library, self.library)